@@ -1,38 +1,26 @@
-from collections import OrderedDict
+import warnings
 from collections.abc import Mapping, Sequence
-from functools import partial
 from ipaddress import ip_address
-from urllib.parse import (SplitResult, parse_qsl,
-                          urljoin, urlsplit, urlunsplit)
+from urllib.parse import SplitResult, parse_qsl, urljoin, urlsplit, urlunsplit
 
 from multidict import MultiDict, MultiDictProxy
+import idna
 
 
-from .quoting import quote, unquote
+from .quoting import _Quoter, _Unquoter
 
-__version__ = '0.10.3'
+__version__ = "1.3.0"
 
-__all__ = ['URL']
+__all__ = ("URL",)
 
 
 # is_leaf()
-# path normalization
 
 
-DEFAULT_PORTS = {
-    'http': 80,
-    'https': 443,
-    'ws': 80,
-    'wss': 443,
-}
-
-PROTECT_CHARS = '=+&;'
+DEFAULT_PORTS = {"http": 80, "https": 443, "ws": 80, "wss": 443}
 
 sentinel = object()
 
-_quote = quote
-_unquote = unquote
-
 
 class cached_property:
     """Use as a class method decorator.  It operates almost exactly like
@@ -47,7 +35,7 @@ class cached_property:
         self.wrapped = wrapped
         try:
             self.__doc__ = wrapped.__doc__
-        except:  # pragma: no cover
+        except AttributeError:  # pragma: no cover
             self.__doc__ = ""
         self.name = wrapped.__name__
 
@@ -135,88 +123,103 @@ class URL:
     #               / path-noscheme
     #               / path-empty
     # absolute-URI  = scheme ":" hier-part [ "?" query ]
-    __slots__ = ('_cache', '_val', '_strict')
+    __slots__ = ("_cache", "_val")
 
-    def __init__(self, val='', *, encoded=False, strict=False):
-        if isinstance(val, URL):
-            self._val = val._val
-            self._cache = val._cache
-            self._strict = val._strict
-            return
-        if isinstance(val, str):
+    _QUOTER = _Quoter()
+    _PATH_QUOTER = _Quoter(safe="@:", protected="/+")
+    _QUERY_QUOTER = _Quoter(safe="?/:@", protected="=+&;", qs=True)
+    _QUERY_PART_QUOTER = _Quoter(safe="?/:@", qs=True)
+    _FRAGMENT_QUOTER = _Quoter(safe="?/:@")
+
+    _UNQUOTER = _Unquoter()
+    _PATH_UNQUOTER = _Unquoter(unsafe="+")
+    _QS_UNQUOTER = _Unquoter(qs=True)
+
+    def __new__(cls, val="", *, encoded=False, strict=None):
+        if strict is not None:  # pragma: no cover
+            warnings.warn("strict parameter is ignored")
+        if type(val) is cls:
+            return val
+        if type(val) is str:
             val = urlsplit(val)
-        elif isinstance(val, SplitResult):
+        elif type(val) is SplitResult:
             if not encoded:
                 raise ValueError("Cannot apply decoding to SplitResult")
+        elif isinstance(val, str):
+            val = urlsplit(str(val))
         else:
             raise TypeError("Constructor parameter should be str")
 
-        self._strict = strict
-
         if not encoded:
             if not val[1]:  # netloc
-                netloc = ''
+                netloc = ""
+                host = ""
             else:
-                netloc = val.hostname
-                if netloc is None:
-                    raise ValueError(
-                        "Invalid URL: host is required for abolute urls.")
+                host = val.hostname
+                if host is None:
+                    raise ValueError("Invalid URL: host is required for abolute urls.")
+
                 try:
-                    netloc.encode('ascii')
-                except UnicodeEncodeError:
-                    netloc = netloc.encode('idna').decode('ascii')
-                else:
-                    try:
-                        ip = ip_address(netloc)
-                    except:
-                        pass
-                    else:
-                        if ip.version == 6:
-                            netloc = '['+netloc+']'
-                if val.port:
-                    netloc += ':{}'.format(val.port)
-                if val.username:
-                    user = _quote(val.username)
-                    if val.password:
-                        user += ':' + _quote(val.password)
-                    netloc = user + '@' + netloc
+                    port = val.port
+                except ValueError:
+                    raise ValueError("Invalid URL: port can't be converted to integer")
 
-            val = SplitResult(
-                val[0],  # scheme
-                netloc,
-                _quote(val[2], safe='+@:', protected='/+', strict=strict),
-                query=_quote(val[3], safe='=+&?/:@',
-                             protected=PROTECT_CHARS, qs=True, strict=strict),
-                fragment=_quote(val[4], safe='?/:@', strict=strict))
+                netloc = cls._make_netloc(
+                    val.username, val.password, host, port, encode=True
+                )
+            path = cls._PATH_QUOTER(val[2])
+            if netloc:
+                path = cls._normalize_path(path)
 
+            cls._validate_authority_uri_abs_path(host=host, path=path)
+            query = cls._QUERY_QUOTER(val[3])
+            fragment = cls._FRAGMENT_QUOTER(val[4])
+            val = SplitResult(val[0], netloc, path, query, fragment)
+
+        self = object.__new__(cls)
         self._val = val
         self._cache = {}
+        return self
 
     @classmethod
-    def build(cls, *, scheme='', user='', password='', host='', port=None,
-              path='', query=None, query_string='', fragment='', strict=False):
+    def build(
+        cls,
+        *,
+        scheme="",
+        user="",
+        password="",
+        host="",
+        port=None,
+        path="",
+        query=None,
+        query_string="",
+        fragment="",
+        encoded=False
+    ):
         """Creates and returns a new URL"""
 
-        if host and not scheme:
-            raise ValueError(
-                'Can\'t build URL with "host" but without "scheme".')
         if not host and scheme:
-            raise ValueError(
-                'Can\'t build URL with "scheme" but without "host".')
+            raise ValueError('Can\'t build URL with "scheme" but without "host".')
+        if port and not host:
+            raise ValueError('Can\'t build URL with "port" but without "host".')
         if query and query_string:
-            raise ValueError(
-                "Only one of \"query\" or \"query_string\" should be passed")
+            raise ValueError('Only one of "query" or "query_string" should be passed')
+
+        if not user and not password and not host and not port:
+            netloc = ""
+        else:
+            netloc = cls._make_netloc(user, password, host, port, encode=not encoded)
+        if not encoded:
+            path = cls._PATH_QUOTER(path)
+            if netloc:
+                path = cls._normalize_path(path)
+
+            cls._validate_authority_uri_abs_path(host=host, path=path)
+            query_string = cls._QUERY_QUOTER(query_string)
+            fragment = cls._FRAGMENT_QUOTER(fragment)
 
         url = cls(
-            SplitResult(
-                scheme,
-                cls._make_netloc(user, password, host, port),
-                _quote(path, safe='@:', protected='/'),
-                _quote(query_string),
-                fragment
-            ),
-            strict=strict,
-            encoded=True
+            SplitResult(scheme, netloc, path, query_string, fragment), encoded=True
         )
 
         if query:
@@ -224,85 +227,94 @@ class URL:
         else:
             return url
 
+    def __init_subclass__(cls):
+        raise TypeError("Inheritance a class {!r} from URL " "is forbidden".format(cls))
+
     def __str__(self):
         val = self._val
         if not val.path and self.is_absolute() and (val.query or val.fragment):
-            val = val._replace(path='/')
+            val = val._replace(path="/")
         return urlunsplit(val)
 
     def __repr__(self):
         return "{}('{}')".format(self.__class__.__name__, str(self))
 
     def __eq__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
 
         val1 = self._val
         if not val1.path and self.is_absolute():
-            val1 = val1._replace(path='/')
+            val1 = val1._replace(path="/")
 
         val2 = other._val
         if not val2.path and other.is_absolute():
-            val2 = val2._replace(path='/')
+            val2 = val2._replace(path="/")
 
         return val1 == val2
 
     def __hash__(self):
-        ret = self._cache.get('hash')
+        ret = self._cache.get("hash")
         if ret is None:
             val = self._val
             if not val.path and self.is_absolute():
-                val = val._replace(path='/')
-            ret = self._cache['hash'] = hash(val)
+                val = val._replace(path="/")
+            ret = self._cache["hash"] = hash(val)
         return ret
 
     def __le__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val <= other._val
 
     def __lt__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val < other._val
 
     def __ge__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val >= other._val
 
     def __gt__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val > other._val
 
     def __truediv__(self, name):
-        name = _quote(name, safe=':@', protected='/', strict=self._strict)
-        if name.startswith('/'):
-            raise ValueError("Appending path "
-                             "starting from slash is forbidden")
+        name = self._PATH_QUOTER(name)
+        if name.startswith("/"):
+            raise ValueError("Appending path " "starting from slash is forbidden")
         path = self._val.path
-        if path == '/':
-            new_path = '/' + name
+        if path == "/":
+            new_path = "/" + name
         elif not path and not self.is_absolute():
             new_path = name
         else:
-            parts = path.rstrip('/').split('/')
+            parts = path.rstrip("/").split("/")
             parts.append(name)
-            new_path = '/'.join(parts)
-        return URL(self._val._replace(path=new_path, query='', fragment=''),
-                   encoded=True)
+            new_path = "/".join(parts)
+        if self.is_absolute():
+            new_path = self._normalize_path(new_path)
+        return URL(
+            self._val._replace(path=new_path, query="", fragment=""), encoded=True
+        )
+
+    def __bool__(self) -> bool:
+        return bool(
+            self._val.netloc or self._val.path or self._val.query or self._val.fragment
+        )
 
     def __getstate__(self):
-        return self._val, self._strict
+        return (self._val,)
 
     def __setstate__(self, state):
         if state[0] is None and isinstance(state[1], dict):
             # default style pickle
-            self._val = state[1]['_val']
-            self._strict = state[1]['_strict']
+            self._val = state[1]["_val"]
         else:
-            self._val, self._strict = state
+            self._val, *unused = state
         self._cache = {}
 
     def is_absolute(self):
@@ -341,8 +353,8 @@ class URL:
         if not self._val.scheme:
             raise ValueError("URL should have scheme")
         v = self._val
-        netloc = self._make_netloc(None, None, v.hostname, v.port)
-        val = v._replace(netloc=netloc, path='', query='', fragment='')
+        netloc = self._make_netloc(None, None, v.hostname, v.port, encode=False)
+        val = v._replace(netloc=netloc, path="", query="", fragment="")
         return URL(val, encoded=True)
 
     def relative(self):
@@ -353,7 +365,7 @@ class URL:
         """
         if not self.is_absolute():
             raise ValueError("URL should be absolute")
-        val = self._val._replace(scheme='', netloc='')
+        val = self._val._replace(scheme="", netloc="")
         return URL(val, encoded=True)
 
     @property
@@ -373,7 +385,10 @@ class URL:
 
         """
         # not .username
-        return self._val.username
+        ret = self._val.username
+        if not ret:
+            return None
+        return ret
 
     @cached_property
     def user(self):
@@ -382,7 +397,7 @@ class URL:
         None if user is missing.
 
         """
-        return _unquote(self.raw_user)
+        return self._UNQUOTER(self.raw_user)
 
     @property
     def raw_password(self):
@@ -400,7 +415,7 @@ class URL:
         None if password is missing.
 
         """
-        return _unquote(self.raw_password)
+        return self._UNQUOTER(self.raw_password)
 
     @property
     def raw_host(self):
@@ -423,11 +438,20 @@ class URL:
         raw = self.raw_host
         if raw is None:
             return None
-        return raw.encode('ascii').decode('idna')
+        if "%" in raw:
+            # Hack for scoped IPv6 addresses like
+            # fe80::2%Проверка
+            # presence of '%' sign means only IPv6 address, so idna is useless.
+            return raw
+
+        try:
+            return idna.decode(raw.encode("ascii"))
+        except UnicodeError:  # e.g. '::1'
+            return raw.encode("ascii").decode("idna")
 
     @property
     def port(self):
-        """Port part of URL.
+        """Port part of URL, with scheme-based fallback.
 
         None for relative URLs or URLs without explicit port and
         scheme without default port substitution.
@@ -435,6 +459,15 @@ class URL:
         """
         return self._val.port or DEFAULT_PORTS.get(self._val.scheme)
 
+    @property
+    def explicit_port(self):
+        """Port part of URL, without scheme-based fallback.
+
+        None for relative URLs or URLs without explicit port.
+
+        """
+        return self._val.port
+
     @property
     def raw_path(self):
         """Encoded path of URL.
@@ -444,7 +477,7 @@ class URL:
         """
         ret = self._val.path
         if not ret and self.is_absolute():
-            ret = '/'
+            ret = "/"
         return ret
 
     @cached_property
@@ -454,7 +487,7 @@ class URL:
         / for absolute URLs without path part.
 
         """
-        return _unquote(self.raw_path, unsafe='+')
+        return self._PATH_UNQUOTER(self.raw_path)
 
     @cached_property
     def query(self):
@@ -483,15 +516,21 @@ class URL:
         Empty string if query is missing.
 
         """
-        return _unquote(self.raw_query_string, qs=True)
+        return self._QS_UNQUOTER(self.raw_query_string)
 
     @cached_property
     def path_qs(self):
-        """Decoded path of URL with query
-        """
+        """Decoded path of URL with query."""
         if not self.query_string:
             return self.path
-        return '{}?{}'.format(self.path, self.query_string)
+        return "{}?{}".format(self.path, self.query_string)
+
+    @cached_property
+    def raw_path_qs(self):
+        """Encoded path of URL with query."""
+        if not self.raw_query_string:
+            return self.raw_path
+        return "{}?{}".format(self.raw_path, self.raw_query_string)
 
     @property
     def raw_fragment(self):
@@ -509,7 +548,7 @@ class URL:
         Empty string if fragment is missing.
 
         """
-        return _unquote(self.raw_fragment)
+        return self._UNQUOTER(self.raw_fragment)
 
     @cached_property
     def raw_parts(self):
@@ -521,14 +560,14 @@ class URL:
         path = self._val.path
         if self.is_absolute():
             if not path:
-                parts = ['/']
+                parts = ["/"]
             else:
-                parts = ['/'] + path[1:].split('/')
+                parts = ["/"] + path[1:].split("/")
         else:
-            if path.startswith('/'):
-                parts = ['/'] + path[1:].split('/')
+            if path.startswith("/"):
+                parts = ["/"] + path[1:].split("/")
             else:
-                parts = path.split('/')
+                parts = path.split("/")
         return tuple(parts)
 
     @cached_property
@@ -538,7 +577,7 @@ class URL:
         ('/',) for absolute URLs if *path* is missing.
 
         """
-        return tuple(_unquote(part) for part in self.raw_parts)
+        return tuple(self._UNQUOTER(part) for part in self.raw_parts)
 
     @cached_property
     def parent(self):
@@ -547,14 +586,12 @@ class URL:
 
         """
         path = self.raw_path
-        if not path or path == '/':
+        if not path or path == "/":
             if self.raw_fragment or self.raw_query_string:
-                return URL(self._val._replace(query='', fragment=''),
-                           encoded=True)
+                return URL(self._val._replace(query="", fragment=""), encoded=True)
             return self
-        parts = path.split('/')
-        val = self._val._replace(path='/'.join(parts[:-1]),
-                                 query='', fragment='')
+        parts = path.split("/")
+        val = self._val._replace(path="/".join(parts[:-1]), query="", fragment="")
         return URL(val, encoded=True)
 
     @cached_property
@@ -564,7 +601,7 @@ class URL:
         if self.is_absolute():
             parts = parts[1:]
             if not parts:
-                return ''
+                return ""
             else:
                 return parts[-1]
         else:
@@ -573,19 +610,87 @@ class URL:
     @cached_property
     def name(self):
         """The last part of parts."""
-        return _unquote(self.raw_name)
+        return self._UNQUOTER(self.raw_name)
+
+    @staticmethod
+    def _validate_authority_uri_abs_path(host, path):
+        """Ensure that path in URL with authority starts with a leading slash.
+
+        Raise ValueError if not.
+        """
+        if len(host) > 0 and len(path) > 0 and not path.startswith("/"):
+            raise ValueError(
+                "Path in a URL with authority " "should start with a slash ('/') if set"
+            )
 
     @classmethod
-    def _make_netloc(cls, user, password, host, port):
-        ret = host
+    def _normalize_path(cls, path):
+        # Drop '.' and '..' from path
+
+        segments = path.split("/")
+        resolved_path = []
+
+        for seg in segments:
+            if seg == "..":
+                try:
+                    resolved_path.pop()
+                except IndexError:
+                    # ignore any .. segments that would otherwise cause an
+                    # IndexError when popped from resolved_path if
+                    # resolving for rfc3986
+                    pass
+            elif seg == ".":
+                continue
+            else:
+                resolved_path.append(seg)
+
+        if segments[-1] in (".", ".."):
+            # do some post-processing here.
+            # if the last segment was a relative dir,
+            # then we need to append the trailing '/'
+            resolved_path.append("")
+
+        return "/".join(resolved_path)
+
+    @classmethod
+    def _encode_host(cls, host):
+        try:
+            ip, sep, zone = host.partition("%")
+            ip = ip_address(ip)
+        except ValueError:
+            try:
+                host = idna.encode(host, uts46=True).decode("ascii")
+            except UnicodeError:
+                host = host.encode("idna").decode("ascii")
+        else:
+            host = ip.compressed
+            if sep:
+                host += "%" + zone
+            if ip.version == 6:
+                host = "[" + host + "]"
+        return host
+
+    @classmethod
+    def _make_netloc(cls, user, password, host, port, encode):
+        if encode:
+            ret = cls._encode_host(host)
+        else:
+            ret = host
         if port:
-            ret = ret + ':' + str(port)
+            ret = ret + ":" + str(port)
         if password:
             if not user:
-                raise ValueError("Non-empty password requires non-empty user")
-            user = user + ':' + password
+                user = ""
+            else:
+                if encode:
+                    user = cls._QUOTER(user)
+            if encode:
+                password = cls._QUOTER(password)
+            user = user + ":" + password
+        elif user and encode:
+            user = cls._QUOTER(user)
         if user:
-            ret = user + '@' + ret
+            ret = user + "@" + ret
         return ret
 
     def with_scheme(self, scheme):
@@ -594,8 +699,7 @@ class URL:
         if not isinstance(scheme, str):
             raise TypeError("Invalid scheme type")
         if not self.is_absolute():
-            raise ValueError("scheme replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("scheme replacement is not allowed " "for relative URLs")
         return URL(self._val._replace(scheme=scheme.lower()), encoded=True)
 
     def with_user(self, user):
@@ -611,18 +715,20 @@ class URL:
         if user is None:
             password = None
         elif isinstance(user, str):
-            user = _quote(user)
+            user = self._QUOTER(user)
             password = val.password
         else:
             raise TypeError("Invalid user type")
         if not self.is_absolute():
-            raise ValueError("user replacement is not allowed "
-                             "for relative URLs")
-        return URL(self._val._replace(netloc=self._make_netloc(user,
-                                                               password,
-                                                               val.hostname,
-                                                               val.port)),
-                   encoded=True)
+            raise ValueError("user replacement is not allowed " "for relative URLs")
+        return URL(
+            self._val._replace(
+                netloc=self._make_netloc(
+                    user, password, val.hostname, val.port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_password(self, password):
         """Return a new URL with password replaced.
@@ -636,20 +742,20 @@ class URL:
         if password is None:
             pass
         elif isinstance(password, str):
-            password = _quote(password)
+            password = self._QUOTER(password)
         else:
             raise TypeError("Invalid password type")
         if not self.is_absolute():
-            raise ValueError("password replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("password replacement is not allowed " "for relative URLs")
         val = self._val
         return URL(
             self._val._replace(
-                netloc=self._make_netloc(val.username,
-                                         password,
-                                         val.hostname,
-                                         val.port)),
-            encoded=True)
+                netloc=self._make_netloc(
+                    val.username, password, val.hostname, val.port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_host(self, host):
         """Return a new URL with host replaced.
@@ -664,24 +770,19 @@ class URL:
         if not isinstance(host, str):
             raise TypeError("Invalid host type")
         if not self.is_absolute():
-            raise ValueError("host replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("host replacement is not allowed " "for relative URLs")
         if not host:
             raise ValueError("host removing is not allowed")
-        try:
-            ip = ip_address(host)
-        except:
-            host = host.encode('idna').decode('ascii')
-        else:
-            if ip.version == 6:
-                host = '['+host+']'
+        host = self._encode_host(host)
         val = self._val
         return URL(
-            self._val._replace(netloc=self._make_netloc(val.username,
-                                                        val.password,
-                                                        host,
-                                                        val.port)),
-            encoded=True)
+            self._val._replace(
+                netloc=self._make_netloc(
+                    val.username, val.password, host, val.port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
     def with_port(self, port):
         """Return a new URL with port replaced.
@@ -691,24 +792,80 @@ class URL:
         """
         # N.B. doesn't cleanup query/fragment
         if port is not None and not isinstance(port, int):
-            raise TypeError(
-                "port should be int or None, got {}".format(type(port)))
+            raise TypeError("port should be int or None, got {}".format(type(port)))
         if not self.is_absolute():
-            raise ValueError("port replacement is not allowed "
-                             "for relative URLs")
+            raise ValueError("port replacement is not allowed " "for relative URLs")
         val = self._val
         return URL(
-            self._val._replace(netloc=self._make_netloc(val.username,
-                                                        val.password,
-                                                        val.hostname,
-                                                        port)),
-            encoded=True)
+            self._val._replace(
+                netloc=self._make_netloc(
+                    val.username, val.password, val.hostname, port, encode=False
+                )
+            ),
+            encoded=True,
+        )
 
-    def with_path(self, path, encoded=False):
+    def with_path(self, path, *, encoded=False):
         """Return a new URL with path replaced."""
         if not encoded:
-            path = _quote(path, safe='@:', protected='/', strict=self._strict)
-        return URL(self._val._replace(path=path), encoded=True)
+            path = self._PATH_QUOTER(path)
+            if self.is_absolute():
+                path = self._normalize_path(path)
+        if len(path) > 0 and path[0] != "/":
+            path = "/" + path
+        return URL(self._val._replace(path=path, query="", fragment=""), encoded=True)
+
+    @staticmethod
+    def _query_var(v):
+        if isinstance(v, str):
+            return v
+        if type(v) is int:  # no subclasses like bool
+            return str(v)
+        raise TypeError(
+            "Invalid variable type: value "
+            "should be str or int, got {!r} "
+            "of type {}".format(v, type(v))
+        )
+
+    def _get_str_query(self, *args, **kwargs):
+        if kwargs:
+            if len(args) > 0:
+                raise ValueError(
+                    "Either kwargs or single query parameter " "must be present"
+                )
+            query = kwargs
+        elif len(args) == 1:
+            query = args[0]
+        else:
+            raise ValueError(
+                "Either kwargs or single query parameter " "must be present"
+            )
+
+        if query is None:
+            query = ""
+        elif isinstance(query, Mapping):
+            quoter = self._QUERY_PART_QUOTER
+            query = "&".join(
+                quoter(k) + "=" + quoter(self._query_var(v)) for k, v in query.items()
+            )
+        elif isinstance(query, str):
+            query = self._QUERY_QUOTER(query)
+        elif isinstance(query, (bytes, bytearray, memoryview)):
+            raise TypeError(
+                "Invalid query type: bytes, bytearray and " "memoryview are forbidden"
+            )
+        elif isinstance(query, Sequence):
+            quoter = self._QUERY_PART_QUOTER
+            query = "&".join(
+                quoter(k) + "=" + quoter(self._query_var(v)) for k, v in query
+            )
+        else:
+            raise TypeError(
+                "Invalid query type: only str, mapping or "
+                "sequence of (str, str) pairs is allowed"
+            )
+
+        return query
 
     def with_query(self, *args, **kwargs):
         """Return a new URL with query part replaced.
@@ -716,6 +873,8 @@ class URL:
         Accepts any Mapping (e.g. dict, multidict.MultiDict instances)
         or str, autoencode the argument if needed.
 
+        A sequence of (key, value) pairs is supported as well.
+
         It also can take an arbitrary number of keyword arguments.
 
         Clear query if None is passed.
@@ -723,80 +882,19 @@ class URL:
         """
         # N.B. doesn't cleanup query/fragment
 
-        if kwargs:
-            if len(args) > 0:
-                raise ValueError("Either kwargs or single query parameter "
-                                 "must be present")
-            query = kwargs
-        elif len(args) == 1:
-            query = args[0]
-        else:
-            raise ValueError("Either kwargs or single query parameter "
-                             "must be present")
-
-        if query is None:
-            query = ''
-        elif isinstance(query, Mapping):
-            quoter = partial(_quote, qs=True, strict=self._strict)
-            lst = []
-            for k, v in query.items():
-                if isinstance(v, str):
-                    pass
-                elif type(v) == int:  # no subclasses like bool
-                    v = str(v)
-                else:
-                    raise TypeError("Invalid variable type: mapping value "
-                                    "should be str or int, got {!r}".format(v))
-                lst.append(quoter(k, safe='/?:@')+'='+quoter(v, safe='/?:@;'))
-                query = '&'.join(lst)
-        elif isinstance(query, str):
-            query = _quote(query, safe='/?:@',
-                           protected=PROTECT_CHARS,
-                           qs=True, strict=self._strict)
-        elif isinstance(query, (bytes, bytearray, memoryview)):
-            raise TypeError("Invalid query type: bytes, bytearray and "
-                            "memoryview are forbidden")
-        elif isinstance(query, Sequence):
-            quoter = partial(_quote, qs=True, strict=self._strict)
-            query = '&'.join(quoter(k, safe='/?:@')+'='+quoter(v, safe='/?:@;')
-                             for k, v in query)
-        else:
-            raise TypeError("Invalid query type: only str, mapping or "
-                            "sequence of (str, str) pairs is allowed")
-        path = self._val.path
-        if path == '':
-            path = '/'
+        new_query = self._get_str_query(*args, **kwargs)
         return URL(
-            self._val._replace(path=path, query=query), encoded=True)
+            self._val._replace(path=self._val.path, query=new_query), encoded=True
+        )
 
     def update_query(self, *args, **kwargs):
         """Return a new URL with query part updated."""
-        if kwargs:
-            if len(args) > 0:
-                raise ValueError("Either kwargs or single query parameter "
-                                 "must be present")
-            new_query = kwargs
-        elif len(args) == 1:
-            new_query = args[0]
-
-            if isinstance(new_query, str):
-                new_query = OrderedDict(
-                    map(
-                        lambda x: x.split('=', 1),
-                        _quote(new_query,
-                               safe='/?:@', protected=PROTECT_CHARS,
-                               qs=True,
-                               strict=self._strict).lstrip("?").split("&")
-                        )
-                )
-
-        else:
-            raise ValueError("Either kwargs or single query parameter "
-                             "must be present")
-
-        query = OrderedDict(self.query)
+        s = self._get_str_query(*args, **kwargs)
+        new_query = MultiDict(parse_qsl(s, keep_blank_values=True))
+        query = MultiDict(self.query)
         query.update(new_query)
-        return self.with_query(tuple(query.items()))
+
+        return URL(self._val._replace(query=self._get_str_query(query)), encoded=True)
 
     def with_fragment(self, fragment):
         """Return a new URL with fragment replaced.
@@ -808,13 +906,12 @@ class URL:
         """
         # N.B. doesn't cleanup query/fragment
         if fragment is None:
-            fragment = ''
+            fragment = ""
         elif not isinstance(fragment, str):
             raise TypeError("Invalid fragment type")
         return URL(
-            self._val._replace(
-                fragment=_quote(fragment, safe='?/:@', strict=self._strict)),
-            encoded=True)
+            self._val._replace(fragment=self._FRAGMENT_QUOTER(fragment)), encoded=True
+        )
 
     def with_name(self, name):
         """Return a new URL with name (last part of path) replaced.
@@ -827,23 +924,26 @@ class URL:
         # N.B. DOES cleanup query/fragment
         if not isinstance(name, str):
             raise TypeError("Invalid name type")
-        if '/' in name:
+        if "/" in name:
             raise ValueError("Slash in name is not allowed")
-        name = _quote(name, safe='@:', protected='/')
+        name = self._PATH_QUOTER(name)
+        if name in (".", ".."):
+            raise ValueError(". and .. values are forbidden")
         parts = list(self.raw_parts)
         if self.is_absolute():
             if len(parts) == 1:
                 parts.append(name)
             else:
                 parts[-1] = name
-            parts[0] = ''  # replace leading '/'
+            parts[0] = ""  # replace leading '/'
         else:
             parts[-1] = name
-            if parts[0] == '/':
-                parts[0] = ''  # replace leading '/'
+            if parts[0] == "/":
+                parts[0] = ""  # replace leading '/'
         return URL(
-            self._val._replace(path='/'.join(parts),
-                               query='', fragment=''), encoded=True)
+            self._val._replace(path="/".join(parts), query="", fragment=""),
+            encoded=True,
+        )
 
     def join(self, url):
         """Join URLs
@@ -865,11 +965,14 @@ class URL:
     def human_repr(self):
         """Return decoded human readable string for URL representation."""
 
-        return urlunsplit(SplitResult(self.scheme,
-                                      self._make_netloc(self.user,
-                                                        self.password,
-                                                        self.host,
-                                                        self._val.port),
-                                      self.path,
-                                      self.query_string,
-                                      self.fragment))
+        return urlunsplit(
+            SplitResult(
+                self.scheme,
+                self._make_netloc(
+                    self.user, self.password, self.host, self._val.port, encode=False
+                ),
+                self.path,
+                self.query_string,
+                self.fragment,
+            )
+        )
